<?php

namespace Drupal\commerce_stripe_payment_request_button\Services;

use Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface;
use Stripe\Stripe;
use Stripe\PaymentIntent;
use Stripe\Exception\ApiErrorException;

use Drupal\Core\Controller\ControllerBase;
use Drupal\commerce_stripe\ErrorHelper;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_shipping\Entity\Shipment;
use Drupal\commerce_shipping\ShipmentItem;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;
use Drupal\physical\Weight;
use Drupal\physical\WeightUnit;
use Drupal\profile\Entity\Profile;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal;
use Drupal\commerce_payment\Entity\PaymentGateway;
use Drupal\commerce_payment\Entity\PaymentGatewayInterface;

use Drupal\commerce_price\Price;
use Exception;

/**
 * Class PaymentRequestButtonHelperService
 *
 * @package Drupal\commerce_stripe_payment_request_button\Service
 */
class PaymentRequestButtonHelperService implements PaymentRequestButtonHelperServiceInterface {
  /**
   * Check if the payment is Google Pay or Apple pay
   *
   * @param PaymentIntent $paymentIntent
   * @return string $wallet
   */
  public function getWalletType(PaymentIntent $paymentIntent): string {
    $wallet = 'Google pay';
    if (isset($paymentIntent->charges->data[0])) {
      /**
       * @var \Stripe\Charge $charges
       */
      $charges = $paymentIntent->charges->data[0];
      $charges_values = $charges->jsonSerialize();
      $wallet = isset($charges_values['payment_method_details']['card']['wallet']['type']) ? 'Apple pay' : 'Google pay';
    }
    return $wallet;
  }

  /**
   * Get the order configuration by currency
   *
   * @param string $currency
   * @return class $data
   */
  public function getOrderPaymentConfiguration(string $currency) {

    $configBase = 'commerce_stripe_payment_request_button';
    $configCurrency = $configBase . '_' . $currency;
    $configName = $configCurrency;

    // Load curency config
    $config = \Drupal::config('commerce_payment.commerce_payment_gateway.' . $configCurrency);

    if (!$config->get('status')) {
      // Load default config
      $config = \Drupal::config('commerce_payment.commerce_payment_gateway.' . $configBase);
      $configName = $configBase;
    }

    $data = new \stdClass();

    $data->config = $config->get('configuration');
    $data->configName = $configName;

    return $data;
  }

  /**
   * Update drupal payment from the payment intent
   *
   * @param OrderInterface $order
   * @param string $intentId
   * @return PaymentIntent $paymentIntent
   */
  public function updateDrupalPaymentByPaymentIntent(OrderInterface $order, string $intentId) {

    $paymentIntent = PaymentIntent::retrieve($intentId);
    $payments = \Drupal::entityQuery('commerce_payment')
      ->condition('order_id', $order->id())
      ->execute();

    $payment = Payment::load(reset($payments));
    $payment->setRemoteId($paymentIntent->id);
    $payment->save();

    // Reset this step as it gets to NULL on cart refresh
    $order->get('checkout_flow')->setValue("payment_request_button_flow");
    $order->get('checkout_step')->setValue("payment");
    $order->save();
    return  $paymentIntent;
  }

  /**
   * Set the following order data: checkout_flow,checkout_step,email,payment_gateway,stripe_intent,wallet
   *
   * @param OrderInterface $order
   * @param PaymentIntent $paymentIntent
   * @param string $email
   * @param string $configName
   * @param string $shipping
   * @return void
   */
  public function setOrderData(OrderInterface $order, PaymentIntent $paymentIntent, string $email, string $configName, string $shipping = ''): void {
    $order->get('checkout_flow')->setValue("payment_request_button_flow");
    $order->get('checkout_step')->setValue("payment");
    $order->setEmail($email);
    $order->get('payment_gateway')->setValue($configName);
    $order->setData('stripe_intent', $paymentIntent->id);
    $order->setData('wallet', $this->getWalletType($paymentIntent));
    if (!empty($shipping)) {
      $order->setData('payment_request_button_shipping', $shipping);
    }
    $order->save();
  }


  /**
   * Create a drupal payment from the data provided
   *
   * @param OrderInterface $order
   * @param PaymentGatewayInterface $paymentGateway
   * @param PaymentIntent $paymentIntent
   * @param string $currentUser
   * @return Payment $payment
   */
  public function createDrupalPayment(OrderInterface $order, PaymentGatewayInterface $paymentGateway, PaymentIntent $paymentIntent, string $currentUser) {
    $payment = Payment::create([
      'state' => 'new',
      'amount' => $order->getTotalPrice(),
      'payment_gateway' => $paymentGateway->id(),
      'order_id' => $order->id(),
      'remote_id' => $paymentIntent->id,
      'payment_gateway_mode' => $paymentGateway->getPlugin()->getMode(),
      'expires' => 0,
      'uid' => $currentUser,
    ]);
    $payment->save();
    return $payment;
  }


  /**
   * Check if the shipping profile exists or create a new one.
   * Afterwards return the profile for processing
   *
   * @param $order
   * @param $shipping_info
   * @param $requestPhone
   * @param $payerPhone
   *
   * @return mixed|null
   */
  public function getOrderShippingProfile($order, $shipping_info, $requestPhone, $payerPhone) {
    // We only get the full name from Apple/Google pay, so we need to "guess" the given
    // name and the family name.
    if (is_array($shipping_info)) {
      $shipping_info = json_decode(json_encode($shipping_info));
    }
    if (!isset($shipping_info->line1) && isset($shipping_info->addressLine[0])) {
      $shipping_info->line1 = $shipping_info->addressLine[0];
    }
    if (!isset($shipping_info->line2) && isset($shipping_info->addressLine[1])) {
      $shipping_info->line2 = $shipping_info->addressLine[1];
    }

    $names = [];

    if (isset($shipping_info->name)) {
      $names = explode(' ', $shipping_info->name);
    } else {
      $names = explode(' ', $shipping_info->recipient);
    }
    if (isset($shipping_info->postalCode)) {
      $shipping_info->postal_code = $shipping_info->postalCode;
    }
    if (isset($shipping_info->region)) {
      $shipping_info->state = $shipping_info->region;
    }

    $given_name = array_shift($names);
    $family_name = implode(' ', $names);
    $shippingProfile = NULL;

    //if the user is authenticated
    if ($order->getCustomerId() != 0) {
      //load all profiles to see if shipping address already exists to prevent database bloat
      $profiles = \Drupal::entityTypeManager()->getStorage('profile')
        ->loadByProperties([
          'uid' => $order->getCustomerId(),
        ]);
      foreach ($profiles as $profile) {
        //check if the address exists
        if (
          $profile->address->country_code == $shipping_info->country &&
          $profile->address->locality == $shipping_info->city &&
          $profile->address->postal_code == $shipping_info->postal_code &&
          $profile->address->address_line1 == $shipping_info->line1 &&
          $profile->address->given_name == $given_name &&
          $profile->address->family_name == $family_name
        ) {
          if ($profile->hasField('field_phone') && $requestPhone) {
            $profile->set('field_phone', $payerPhone);
          }
          $profile->save();
          $shippingProfile = $profile;
          break;
        }
      }
    }
    $orderShipments = $order->shipments->referencedEntities();

    $firstShipment = reset($orderShipments);

    //if profile exists set it as the order shipping profile else create new order
    if (!is_null($shippingProfile)) {
      $firstShipment->setShippingProfile($shippingProfile);
    } else {
      $shippingProfile = Profile::create([
        'type' => 'customer',
        'uid' => $order->getCustomerId(),
      ]);
      $shippingProfile->save();

      $shippingProfile->address->given_name = $given_name;
      $shippingProfile->address->family_name = $family_name;
      $shippingProfile->address->country_code = $shipping_info->country;
      $shippingProfile->address->locality = $shipping_info->city;
      $shippingProfile->address->administrative_area = $shipping_info->state;
      $shippingProfile->address->postal_code = $shipping_info->postal_code;
      $shippingProfile->address->address_line1 = $shipping_info->line1;
      if (isset($shipping_info->line2)) {
        $shippingProfile->address->address_line2 = $shipping_info->line2;
      }
      if ($shippingProfile->hasField('field_phone') && $requestPhone) {
        $shippingProfile->set('field_phone', $payerPhone);
      }
      $shippingProfile->save();
    }
    return $shippingProfile;
  }

  public function handleShipping($order, $shipping_info, $values, $requestPhone = FALSE) {
    $shipment = null;
    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty()) {
      $shipment = $order->get('shipments')->entity;
    } else {
      $shipment = Shipment::create([
        'type' => 'default',
        'order_id' => $order->id(),
        'title' => 'Shipment',
        'state' => 'ready',
      ]);
    }

    // Loop through order items and add them to shipment
    foreach ($order->getItems() as $order_item) {
      $quantity = $order_item->getQuantity();
      $purchased_entity = $order_item->getPurchasedEntity();

      //TODO Skip weight if there is none, currently breaks order flow
      if ($purchased_entity->get('weight')->isEmpty()) {
        $weight = new Weight(1, WeightUnit::GRAM);
      } else {
        $weight_item = $purchased_entity->get('weight')->first();
        $weight = $weight_item->toMeasurement();
      }

      $shipment_item = new ShipmentItem([
        'order_item_id' => $order_item->id(),
        'title' => $purchased_entity->label(),
        'quantity' => $quantity,
        'weight' => $weight->multiply($quantity),
        'declared_value' => $order_item->getTotalPrice(),
      ]);
      $shipment->addItem($shipment_item);
    }

    $shipping_method = $order->get('shipments')->entity->getShippingMethod();

    if (!is_null($order->getData('payment_request_button_shipping'))) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($order->getData('payment_request_button_shipping'));
    } elseif (isset($values['result']['shippingOption']) &&  $shipping_method->id() != $values['data']['result']['shippingOption']['id']) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($values['data']['result']['shippingOption']['id']);
    }
    $shipment->setTitle($shipping_method->getName());
    $shipment->setShippingMethodId($shipping_method->id());
    $shipment->setShippingMethod($shipping_method);
    $shipment->save();
    $order->set('shipments', $shipment);

    $order->save();

    // Add proper shipping method
    $shipping_method_plugin = $shipping_method->getPlugin();
    $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
    $shipping_rate = reset($shipping_rates);
    $shipping_service = $shipping_rate->getService()->getId();
    $shipping_method_amount = $shipping_rate->getAmount();

    //if the order currency is different from the default currency, check the if there is an correct shipping field set
    if (
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
      ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
        ->getCurrencyCode()])
    ) {
      $newAmount = $shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
        ->getCurrencyCode()];
      if (is_numeric($newAmount['number']) && $newAmount['number'] > 0) {
        $shipping_method_amount = new Price($newAmount['number'], $newAmount['currency_code']);
      }
    }
    //amount is not zero and there are not currency fields appended We use the commerce_currency_resolver to convert the amount
    elseif (
      $shipping_method_amount->getNumber() > 0 &&
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
      ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
        ->getCurrencyCode()])
    ) {
      $shipping_method_amount = Drupal::service('commerce_currency_resolver.calculator')
        ->priceConversion($shipping_method_amount, $order->getTotalPrice()
          ->getCurrencyCode());
    }

    $shipment->setAmount($shipping_method_amount);

    $phone = '';
    if (is_array($values) && isset($values['data']['result']['shippingAddress']['phone'])) {
      $phone = $values['data']['result']['shippingAddress']['phone'];
    } elseif (!is_array($values)) { //just to check that it is an object
      $phone = $values->billing_details->phone;
    }
    $profile = $this->getOrderShippingProfile($order, $shipping_info, $requestPhone, $phone);

    $shipment->setShippingProfile($profile);
    $shipment->setShippingService($shipping_service);
    $shipment->save();

    $order->set('shipments', $shipment);
    $order->save();
  }

  public function processProfile(OrderInterface $order, $paymentCharge, $requestPhone = true) {
    // Deal with billing profile
    $profile = $order->getBillingProfile();
    $currentUser = $order->getCustomerId();

    if (empty($profile)) {
      $profile = Profile::create([
        'type' => 'customer',
        'uid' => $currentUser,
      ]);
      $profile->save();
    }

    $current_name = $paymentCharge->billing_details->name;
    $names = explode(' ', $current_name);
    $given_name = array_shift($names);
    $family_name = implode(' ', $names);

    $profile->address->given_name = $given_name;
    $profile->address->family_name = $family_name;
    $profile->address->country_code = $paymentCharge->billing_details->address->country;
    $profile->address->locality = $paymentCharge->billing_details->address->city;
    $profile->address->administrative_area = $paymentCharge->billing_details->address->state;
    $profile->address->postal_code = $paymentCharge->billing_details->address->postal_code;
    $profile->address->address_line1 = $paymentCharge->billing_details->address->line1;
    if (isset($paymentCharge->billing_details->address->line2)) {
      $profile->address->address_line2 = $paymentCharge->billing_details->address->line2;
    }
    if ($profile->hasField('field_phone') && $requestPhone) {
      $profile->set('field_phone', $paymentCharge->billing_details->phone);
    }
    $profile->save();

    $order->setBillingProfile($profile);
    $order->save();
  }
}
